from setuptools import setup

setup(
    name='pybezier',
    version='0.1.0',
    description='A simple Python package to manipulate Bézier curves',
    url='https://github.com/TobiaMarcucci/pybezier',
    author='Tobia Marcucci',
    author_email='tobiam@mit.edu',
    license='MIT',
    packages=['pybezier'],
    install_requires=['numpy'],
)
